package org.grouplens.common.slf4j.maven;

/**
 * Reference that intelligently inherits from another object.
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
public class InheritableReference<T> {
    private volatile T object;
    private volatile T defaultValue;
    private InheritableReference<T> parent;
    
    /**
     * Create a new null, parentless inheritable reference.
     */
    public InheritableReference() {
        object = null;
        parent = null;
    }
    
    /**
     * Create a parentless reference to the specified object without a parent.
     * @param obj
     */
    public InheritableReference(T obj) {
        this.object = obj;
    }
    
    /**
     * Create an inherited reference.
     * @param parent
     */
    public InheritableReference(InheritableReference<T> parent) {
        this.parent = parent;
    }
    
    T get(boolean useDefault) {
        if (object != null) return object;
        
        T obj;
        if (parent != null && (obj = parent.get(false)) != null)
            return obj;
        
        if (useDefault) {
            if (defaultValue != null)
                return defaultValue;
            else if (parent != null)
                return parent.get(true);
        }
        
        return null;
    }
    
    /**
     * Get the object, consulting the parent if the reference is null.
     * @return
     */
    public T get() {
        return get(true);
    }
    
    /**
     * Set the object pointed to by this reference, overriding any parent object.
     * Setting it to null reverts to using the parent reference.
     * @param obj The object the reference should point to, or <tt>null</tt> to
     * clear the reference.
     */
    public void set(T obj) {
        object = obj;
    }
    
    /**
     * Set a default value to be overridden by parent values.
     */
    public void setDefault(T dft) {
        defaultValue = dft;
    }
    
    /**
     * A thread-local that uses inheritable references to allow overriding.
     * @author Michael Ekstrand <ekstrand@cs.umn.edu>
     *
     */
    public static class ThreadLocal<T> extends InheritableThreadLocal<InheritableReference<T>> {
        protected InheritableReference<T> initialValue() {
            return new InheritableReference<T>();
        }
        protected InheritableReference<T> childValue(InheritableReference<T> parent) {
            return new InheritableReference<T>(parent);
        }
    }
}
